/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

mod utils;

use opentally::election::Election;
use opentally::numbers::Rational;
use opentally::stv;

use csv::StringRecord;
use flate2::bufread::GzDecoder;
use utf8_chars::BufReadCharsExt;

use std::fs::File;
use std::io::{self, BufReader};

#[test]
fn aec_tas19_rational() {
	// Read CSV file
	let reader = csv::ReaderBuilder::new()
		.has_headers(false)
		.from_path("tests/data/aec-senate-formalpreferences-24310-TAS.csv")
		.expect("IO Error");
	let records: Vec<StringRecord> = reader.into_records().map(|r| r.expect("Syntax Error")).collect();
	
	let mut candidates: Vec<&str> = records.iter().skip(2).map(|r| &r[0]).collect();
	// Remove exhausted/LBF rows
	candidates.truncate(candidates.len() - 2);
	
	let stages: Vec<usize> = records.first().unwrap().iter().skip(1).step_by(2).map(|s| s.parse().unwrap()).collect();
	
	// Decompress BLT
	let file = File::open("tests/data/aec-senate-formalpreferences-24310-TAS.blt.gz").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let gz_decoder = GzDecoder::new(file_reader);
	
	let mut reader = BufReader::new(gz_decoder);
	let chars = reader.chars().map(|r| r.expect("IO Error")).peekable();
	
	// Read BLT
	let election: Election<Rational> = Election::from_blt(chars).expect("Syntax Error");
	
	// Validate candidate names
	for (i, candidate) in candidates.iter().enumerate() {
		assert_eq!(election.candidates[i].name, *candidate);
	}
	
	let stv_opts = stv::STVOptions {
		round_tvs: None,
		round_weights: None,
		round_votes: Some(0),
		round_quota: Some(0),
		sum_surplus_transfers: stv::SumSurplusTransfersMode::SingleStep,
		meek_surplus_tolerance: String::new(),
		normalise_ballots: false,
		quota: stv::QuotaType::Droop,
		quota_criterion: stv::QuotaCriterion::GreaterOrEqual,
		quota_mode: stv::QuotaMode::Static,
		ties: vec![],
		surplus: stv::SurplusMethod::UIG,
		surplus_order: stv::SurplusOrder::ByOrder,
		transferable_only: false,
		exclusion: stv::ExclusionMethod::ByValue,
		meek_nz_exclusion: false,
		early_bulk_elect: true,
		bulk_exclude: true,
		defer_surpluses: false,
		meek_immediate_elect: false,
		constraints_path: None,
		constraint_mode: stv::ConstraintMode::GuardDoom,
		hide_excluded: false,
		sort_votes: false,
		pp_decimals: 2,
	};
	utils::validate_election::<Rational>(stages, records, election, stv_opts, None, &["exhausted", "lbf"]);
}
